datacache <- new.env(hash=TRUE, parent=emptyenv())

targetscan.Mm.eg <- function() showQCData("targetscan.Mm.eg", datacache)
targetscan.Mm.eg_dbconn <- function() dbconn(datacache)
targetscan.Mm.eg_dbfile <- function() dbfile(datacache)
targetscan.Mm.eg_dbschema <- function(file="", show.indices=FALSE) dbschema(datacache, file=file, show.indices=show.indices)
targetscan.Mm.eg_dbInfo <- function() dbInfo(datacache)

.onLoad <- function(libname, pkgname)
{
    require("methods", quietly=TRUE)

    setClass("miRNAAnnDbBimap", contains="AnnDbBimap")
    setClass("miRNATargetAnnDbBimap", contains="AnnDbBimap")

    setMethod("as.list", "miRNATargetAnnDbBimap",
              function(x, ...)
              {
                y <- AnnotationDbi:::flatten(x, fromKeys.only=TRUE)
                makemiRNATargetNode <- function(name, UTR_start, UTR_end,
                                                MSA_start, MSA_end, Seed_match, PCT, ...)
                  {
                    l <- mapply(list, SIMPLIFY=FALSE,
                                miR.Family=name,
                                UTR.start=UTR_start,
                                UTR.end=UTR_end,
                                MSA.start=MSA_start,
                                MSA.end=MSA_end,
                                Seed.match=Seed_match,
                                PCT=PCT)
                    for (i in seq_along(l)) { class(l[[i]]) <- "targetscan.Target" }
                    l
                  }
                AnnotationDbi:::.toListOfLists(y, mode=1, makemiRNATargetNode)
              }
              )

    setMethod("as.list", "miRNAAnnDbBimap",
              function(x, ...)
              {
                y <- AnnotationDbi:::flatten(x, fromKeys.only=TRUE)
                makemiRNANode <- function(mirbase_id, MiRBase_Accession,
                                          Seed_m8, Species_ID, Mature_sequence,
                                          Family_conservation, ...)
                  {
                    l <- list(MiRBase.ID=mirbase_id,
                              MiRBase.Accession=MiRBase_Accession,
                              Seed.m8=Seed_m8,
                              Species=Species_ID,
                              Mature.sequence=Mature_sequence,
                              Family.conservation=Family_conservation,
                              ...)
                    class(l) <- "targetscan.MiRBase"
                    l
                  }
                AnnotationDbi:::.toListOfLists(y, mode=1, makemiRNANode)
              }
              )
    
    ## Connect to the SQLite DB
    dbfile <- system.file("extdata", "targetscan.Mm.eg.sqlite", package=pkgname, lib.loc=libname)
    assign("dbfile", dbfile, envir=datacache)
    dbconn <- dbFileConnect(dbfile)
    assign("dbconn", dbconn, envir=datacache)
    ## Create the AnnObj instances
    ann_objs <- createAnnObjs.TARGETSCAN_DB("targetscan.Mm.eg", "TargetScan (Mouse)", dbconn, datacache)
    mergeToNamespaceAndExport(ann_objs, pkgname)
}

.onUnload <- function(libpath)
{
    dbFileDisconnect(targetscan.Mm.eg_dbconn())
}

### Mandatory fields: objName, Class and L2Rchain
TARGETSCAN_DB_AnnDbBimap_seeds <- list(
    list(objName="MIRBASE2FAMILY",
         Class="AnnDbBimap",
         L2Rchain=list(
           list(tablename="mirbase",
                Lcolname="mirbase_id",
                Rcolname="family"
                ),
           list(
                tablename="mirna_family",
                Lcolname="_id",
                Rcolname="name"
                )
           )
         ),
     list(objName="MIRNA",
          Class="miRNAAnnDbBimap",
          L2Rchain=list(
            list(tablename="mirbase",
                 Lcolname="mirbase_id",
                 Rcolname="mirbase_id",
                 Rattribnames=c(
                   MiRBase_Accession="{mirbase_accession}",
                   Seed_m8="{seed_m8}",
                   Species_ID="species.name",
                   Mature_sequence="{mature_sequence}",
                   Family_conservation="{family_conservation}"),
                 Rattrib_join="LEFT JOIN species ON {species}=species.id"
                 )
            )
          ),
      list(objName="TARGETS",
           Class="AnnDbBimap",
           L2Rchain=list(
             list(tablename="genes",
                  Lcolname="gene_id",
                  Rcolname="_id"
                  ),
             list(tablename="targets",
                  Lcolname="target",
                  Rcolname="family"
                  ),
             list(tablename="mirna_family",
                  Lcolname="_id",
                  Rcolname="name"
                  )
             )
           ),
      list(objName="TARGETSFULL",
           Class="miRNATargetAnnDbBimap",
           L2Rchain=list(
             list(tablename="genes",
                  Lcolname="gene_id",
                  Rcolname="_id"
                  ),
             list(tablename="targets",
                  Lcolname="target",
                  Rattribnames=c(
                    UTR_start="{utr_start}",
                    UTR_end="{utr_end}",
                    MSA_start="{msa_start}",
                    MSA_end="{msa_end}",
                    Seed_match="seed_match.name",
                    PCT="{pct}"),
                  Rattrib_join="LEFT JOIN seed_match ON {seed_match}=seed_match._id LEFT JOIN mirna_family AS _R ON {family}=_R._id",
                  Rcolname="name"
##                   ),
##              list(tablename="mirna_family",
##                   Lcolname="_id",
##                   Rcolname="name"
                  )
             )
           )
)

createAnnObjs.TARGETSCAN_DB <- function(prefix, objTarget, dbconn, datacache)
{
    ## checkDBSCHEMA(dbconn, "TARGETSCAN_DB")

    ## AnnDbBimap objects
    seed0 <- list(
        objTarget=objTarget,
        datacache=datacache
    )
    ann_objs <- AnnotationDbi:::createAnnDbBimaps(TARGETSCAN_DB_AnnDbBimap_seeds, seed0)

    ## Reverse maps
    revmap2 <- function(from, to)
    {
        map <- revmap(ann_objs[[from]], objName=to)
        L2Rchain <- map@L2Rchain
        tmp <- L2Rchain[[1]]@filter
        L2Rchain[[1]]@filter <-
          L2Rchain[[length(L2Rchain)]]@filter
        L2Rchain[[length(L2Rchain)]]@filter <- tmp
        map@L2Rchain <- L2Rchain
        map
    }
    ann_objs$FAMILY2MIRBASE <- revmap2("MIRBASE2FAMILY", "FAMILY2MIRBASE")
    
    ## 1 special map that is not an AnnDbBimap object (just a named integer vector)
    ann_objs$MAPCOUNTS <- AnnotationDbi:::createMAPCOUNTS(dbconn, prefix)

    AnnotationDbi:::prefixAnnObjNames(ann_objs, prefix)
}

